/*****************************************************************************/
/*                                                                           */
/* Author: Michael C. Nechyba                                                */
/* Last modified: 06/04/2003                                                 */
/*                                                                           */
/* File: "blob_detect.c"                                                     */
/*                                                                           */
/*****************************************************************************/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#ifdef INCLUDE_JAVA
#include "GatorCameraClient.h"
#endif

#include "acticast_include.h"

/****************************** Subroutines **********************************/

int new_detect_blobs_ppm(RAW *frame, PPM *t, BLOB2 *b, 
			 int min_blob_size)
{
  int i, j, k, count, label;
  int tmp_x, tmp_y, tmp_n;
  int status, n_stack, flag; //  recursive call limit flags
  int flag2, tmp_label, tmp_count;

  count = 0;
  label = 1;

  limit_dimensions(t);

  for(i = 0; i < t->ydim; i++) {
    for(j = 0; j < t->xdim; j++) {
      if(t->p[j][i][0] == (char) 255) { 
	flag2 = check_neighbors_blob(t, j, i, label);
	if(flag2 == 0) {
	  initialize_blob2(&b[count]); 
	  b[count].first.x = j; 
	  b[count].first.y = i;
	  b[count].label = label;
	} else {
	  tmp_label = label;
	  tmp_count = count;
	  label = flag2;
	  count = label - 1;
	}	  

	flag = 0;
	n_stack = 0;
	status = 1;
	tmp_x = j; tmp_y = i;

	while(flag == 0) {
	  new_collect_blob_data_ppm(frame, t, tmp_x, tmp_y, 
				    label, 255, &b[count], 
				    &status, &n_stack);
	  if(status == 0) {
	    b[count].n--;
	    tmp_n = b[count].n;
	    tmp_x = (int) (b[count].xy_t[0][tmp_n] + 0.5);
	    tmp_y = (int) (b[count].xy_t[1][tmp_n] + 0.5);
	    b[count].ell.mean[0] -= tmp_x;
	    b[count].ell.mean[1] -= tmp_y;
	    b[count].ell.cov[0][0] -= (tmp_x*tmp_x);
	    b[count].ell.cov[0][1] -= (tmp_y*tmp_x);
	    b[count].ell.cov[1][0] = b[count].ell.cov[0][1];
	    b[count].ell.cov[1][1] -= (tmp_y*tmp_y);

	    n_stack = 0;
	    status = 1;
	  } else {
	    n_stack = 0;
	    flag = 1;
	  }
	}
	
	if(flag2 == 0) {
	  if(b[count].n >= min_blob_size) {
	    count++;
	    label++;
	  } else {
	    new_collect_blob_data_ppm(frame, t, 
				      b[count].first.x, b[count].first.y,
				      255, label, &b[count],
				      &status, &n_stack);
	  }
	} else {
	  count = tmp_count;
	  label = tmp_label;
	}
	
	if(label >= 255) {
	  fprintf(stderr, "Found more than 254 blobs of size > %d.\n",
		  min_blob_size);
	  fprintf(stderr, "This is a FATAL error.\n");
	  exit(-1);
	}
      }
    }
  }
  
  for(i = 0; i < count; i++) {
    compute_ellipse_from_online(&b[i].ell, b[i].n);
  }
  
#ifdef COLOR_DIAGNOSTIC
  for(i = 0; i < count; i++)
    for(j = 0; j < b[i].n; j++) 
    fprintf(stdout, "RGB: %d %.0f %.0f %.0f %.0f %.0f\n", 
            b[i].n,
            b[i].rgb_t[0][j],
            b[i].rgb_t[1][j],
            b[i].rgb_t[2][j],
	    b[i].xy_t[0][j],
	    b[i].xy_t[1][j]);
#endif

  return(count);
}

/*****************************************************************************/

int new_collect_blob_data_ppm(RAW *frame, PPM *t, 
			      int x, int y, int label, int tst, BLOB2 *b,
			      int *status, int *n_stack)
{
  int loc;
  double xd, yd;
  
  *n_stack += 1;

  if(*n_stack >= MAX_RECURSIONS) {
    *status = 0;
    return(0);
  }

  xd = (double) x;
  yd = (double) y;

  loc = (y*frame->xdim + x)*4;

  t->p[x][y][0] = (char) label;
  b->rgb_t[RED][b->n] =   (double) convert_char(frame->p[loc+RED_RAW]);
  b->rgb_t[GREEN][b->n] = (double) convert_char(frame->p[loc+GREEN_RAW]);
  b->rgb_t[BLUE][b->n] =  (double) convert_char(frame->p[loc+BLUE_RAW]);
  b->xy_t[0][b->n] = xd;
  b->xy_t[1][b->n] = yd;
  b->type[b->n] = 0;
  b->n++;

  b->ell.mean[0] += xd;
  b->ell.mean[1] += yd;
  b->ell.cov[0][0] += (xd*xd);
  b->ell.cov[0][1] += (yd*xd);
  b->ell.cov[1][0] = b->ell.cov[0][1];
  b->ell.cov[1][1] += (yd*yd);

  b->ell.xmin = min_int(b->ell.xmin, x);
  b->ell.xmax = max_int(b->ell.xmax, x);
  b->ell.ymin = min_int(b->ell.ymin, y);
  b->ell.ymax = max_int(b->ell.ymax, y);

#ifdef EIGHT_CONNECTED  
  if((x-1) >= 0 && (y-1) >= 0) 
    if(t->p[x-1][y-1][0] == (char) tst) 
      new_collect_blob_data_ppm(frame, t, x-1, y-1, label, tst, b, 
				status, n_stack);
#endif

  if((y-1) >= 0)
    if(t->p[x][y-1][0] == (char) tst) 
      new_collect_blob_data_ppm(frame, t, x, y-1, label, tst, b, 
				status, n_stack);

#ifdef EIGHT_CONNECTED
  if((x+1) < t->xdim && (y-1) >= 0) 
    if(t->p[x+1][y-1][0] == (char) tst) 
      new_collect_blob_data_ppm(frame, t, x+1, y-1, label, tst, b, 
				status, n_stack);
#endif

  if((x-1) >= 0) 
    if(t->p[x-1][y][0] == (char) tst) 
      new_collect_blob_data_ppm(frame, t, x-1, y, label, tst, b, 
				status, n_stack);

  if((x+1) < t->xdim) 
    if(t->p[x+1][y][0] == (char) tst) 
      new_collect_blob_data_ppm(frame, t, x+1, y, label, tst, b, 
				status, n_stack);

#ifdef EIGHT_CONNECTED
  if((x-1) >= 0 && (y+1) < t->ydim) 
    if(t->p[x-1][y+1][0] == (char) tst) 
      new_collect_blob_data_ppm(frame, t, x-1, y+1, label, tst, b, 
				status, n_stack);
#endif

  if((y+1) < t->ydim) 
    if(t->p[x][y+1][0] == (char) tst) 
      new_collect_blob_data_ppm(frame, t, x, y+1, label, tst, b, 
				status, n_stack);

#ifdef EIGHT_CONNECTED
  if((x+1) < t->xdim && (y+1) < t->ydim)
    if(t->p[x+1][y+1][0] == (char) tst) 
     new_collect_blob_data_ppm(frame, t, x+1, y+1, label, tst, b, 
			       status, n_stack);
#endif

  return(1);
}

/*****************************************************************************/

// FIX VERIFY FIX

int new_detect_blobs_ppm2(RAW2 *frame, PPM2 *t, BLOB2 *b, 
			  int min_blob_size)
{
  int i, j, j1, count, label;
  int tmp_x, tmp_y, tmp_n;
  int status, n_stack, flag; //  recursive call limit flags
  int flag2, tmp_label, tmp_count;

  count = 0;
  label = 1;

  limit_dimensions(&t->im);

  for(i = 0; i < t->im.ydim; i++) {
    for(j1 = 0; j1 < t->m.line[i].n; j1+=2) {      
      for(j = t->m.line[i].c[j1].x; j <= t->m.line[i].c[j1+1].x; j++) {
	if(t->im.p[j][i][0] == (char) 255) { 
	  flag2 = check_neighbors_blob(&t->im, j, i, label);
	  if(flag2 == 0) {
	    initialize_blob2(&b[count]); 
	    b[count].first.x = j; 
	    b[count].first.y = i;
	    b[count].label = label;
	  } else {
	    tmp_label = label;
	    tmp_count = count;
	    label = flag2;
	    count = label - 1;
	  }

	  flag = 0;
	  n_stack = 0;
	  status = 1;
	  tmp_x = j; tmp_y = i;

	  while(flag == 0) {
	    new_collect_blob_data_ppm2(frame, t, tmp_x, tmp_y, 
				       label, 255, &b[count], 
				       &status, &n_stack);
	    if(status == 0) {
	      b[count].n--;
	      tmp_n = b[count].n;
	      tmp_x = (int) (b[count].xy_t[0][tmp_n] + 0.5);
	      tmp_y = (int) (b[count].xy_t[1][tmp_n] + 0.5);

	      b[count].ell.mean[0] -= tmp_x;
	      b[count].ell.mean[1] -= tmp_y;
	      b[count].ell.cov[0][0] -= (tmp_x*tmp_x);
	      b[count].ell.cov[0][1] -= (tmp_y*tmp_x);
	      b[count].ell.cov[1][0] = b[count].ell.cov[0][1];
	      b[count].ell.cov[1][1] -= (tmp_y*tmp_y);
	      
	      n_stack = 0;
	      status = 1;
	    } else {
	      n_stack = 0;
	      flag = 1;
	    }
	  }
	  
	  if(flag2 == 0) {
	    if(b[count].n >= min_blob_size) {
	      count++;
	      label++;
	    } else {
	      new_collect_blob_data_ppm2(frame, t, 
					 b[count].first.x, b[count].first.y,
					 255, label, &b[count],
					 &status, &n_stack);
	    }
	  } else {
	    count = tmp_count;
	    label = tmp_label;
	  }

	  if(label >= 255) {
	    fprintf(stderr, "Found more than 254 blobs of size > %d.\n",
		    min_blob_size);
	    fprintf(stderr, "This is a FATAL error.\n");
	    exit(-1);
	  }
	}
      }
    }
  }
  
  for(i = 0; i < count; i++) {
    compute_ellipse_from_online(&b[i].ell, b[i].n);
  }

#ifdef COLOR_DIAGNOSTIC
  for(i = 0; i < count; i++)
    for(j = 0; j < b[i].n; j++) 
    fprintf(stdout, "RGB: %d %.0f %.0f %.0f %.0f %.0f\n", 
            b[i].n,
            b[i].rgb_t[0][j],
            b[i].rgb_t[1][j],
            b[i].rgb_t[2][j],
	    b[i].xy_t[0][j],
	    b[i].xy_t[1][j]);
#endif

  return(count);
}

/*****************************************************************************/

/*****************************************************************************/

int new_collect_blob_data_ppm2(RAW2 *frame, PPM2 *t, 
			       int x, int y, int label, int tst, BLOB2 *b,
			       int *status, int *n_stack)
{
  int loc;
  double xd, yd;
  
  *n_stack += 1;

  if(*n_stack >= MAX_RECURSIONS) {
    *status = 0;
    return(0);
  }

  xd = (double) x;
  yd = (double) y;

  loc = (y*frame->im.xdim + x)*4;

  t->im.p[x][y][0] = (char) label;
  b->rgb_t[RED][b->n] =   (double) convert_char(frame->im.p[loc+RED_RAW]);
  b->rgb_t[GREEN][b->n] = (double) convert_char(frame->im.p[loc+GREEN_RAW]);
  b->rgb_t[BLUE][b->n] =  (double) convert_char(frame->im.p[loc+BLUE_RAW]);
  b->xy_t[0][b->n] = xd;
  b->xy_t[1][b->n] = yd;
  b->type[b->n] = 0;
  b->n++;

  b->ell.mean[0] += xd;
  b->ell.mean[1] += yd;
  b->ell.cov[0][0] += (xd*xd);
  b->ell.cov[0][1] += (yd*xd);
  b->ell.cov[1][0] = b->ell.cov[0][1];
  b->ell.cov[1][1] += (yd*yd);

  b->ell.xmin = min_int(b->ell.xmin, x);
  b->ell.xmax = max_int(b->ell.xmax, x);
  b->ell.ymin = min_int(b->ell.ymin, y);
  b->ell.ymax = max_int(b->ell.ymax, y);

#ifdef EIGHT_CONNECTED  
  if((x-1) >= 0 && (y-1) >= 0) 
    if(t->im.p[x-1][y-1][0] == (char) tst) 
      new_collect_blob_data_ppm2(frame, t, x-1, y-1, label, tst, b, 
				 status, n_stack);
#endif

  if((y-1) >= 0)
    if(t->im.p[x][y-1][0] == (char) tst) 
      new_collect_blob_data_ppm2(frame, t, x, y-1, label, tst, b, 
				 status, n_stack);

#ifdef EIGHT_CONNECTED
  if((x+1) < t->im.xdim && (y-1) >= 0) 
    if(t->im.p[x+1][y-1][0] == (char) tst) 
      new_collect_blob_data_ppm2(frame, t, x+1, y-1, label, tst, b, 
				 status, n_stack);
#endif

  if((x-1) >= 0) 
    if(t->im.p[x-1][y][0] == (char) tst) 
      new_collect_blob_data_ppm2(frame, t, x-1, y, label, tst, b, 
				 status, n_stack);

  if((x+1) < t->im.xdim) 
    if(t->im.p[x+1][y][0] == (char) tst) 
      new_collect_blob_data_ppm2(frame, t, x+1, y, label, tst, b, 
				 status, n_stack);

#ifdef EIGHT_CONNECTED
  if((x-1) >= 0 && (y+1) < t->im.ydim) 
    if(t->im.p[x-1][y+1][0] == (char) tst) 
      new_collect_blob_data_ppm2(frame, t, x-1, y+1, label, tst, b, 
				 status, n_stack);
#endif

  if((y+1) < t->im.ydim) 
    if(t->im.p[x][y+1][0] == (char) tst) 
      new_collect_blob_data_ppm2(frame, t, x, y+1, label, tst, b, 
				 status, n_stack);

#ifdef EIGHT_CONNECTED
  if((x+1) < t->im.xdim && (y+1) < t->im.ydim)
    if(t->im.p[x+1][y+1][0] == (char) tst) 
     new_collect_blob_data_ppm2(frame, t, x+1, y+1, label, tst, b, 
				status, n_stack);
#endif

  return(1);
}

/*****************************************************************************/

int check_neighbors_blob(PPM *t, int x, int y, int label)
{
  int i, tst;

  //fprintf(stderr, "in check_neighbors_blob()\n");

  for(i = 1; i < label; i++) {
    tst = i;

#ifdef EIGHT_CONNECTED  
    if((x-1) >= 0 && (y-1) >= 0) 
      if(t->p[x-1][y-1][0] == (char) tst) 
	return(i);
#endif

    if((y-1) >= 0)
      if(t->p[x][y-1][0] == (char) tst) 
	return(i);
    
#ifdef EIGHT_CONNECTED
    if((x+1) < t->xdim && (y-1) >= 0) 
      if(t->p[x+1][y-1][0] == (char) tst) 
	return(i);
#endif

    if((x-1) >= 0) 
      if(t->p[x-1][y][0] == (char) tst) 
	return(i);
    
    if((x+1) < t->xdim) 
      if(t->p[x+1][y][0] == (char) tst) 
	return(i);

#ifdef EIGHT_CONNECTED
    if((x-1) >= 0 && (y+1) < t->ydim) 
      if(t->p[x-1][y+1][0] == (char) tst) 
	return(i);
#endif

    if((y+1) < t->ydim) 
      if(t->p[x][y+1][0] == (char) tst) 
	return(i);
    
#ifdef EIGHT_CONNECTED
    if((x+1) < t->xdim && (y+1) < t->ydim)
      if(t->p[x+1][y+1][0] == (char) tst) 
	return(i);
#endif
  }

  return(0);
}
