/*****************************************************************************/
/*                                                                           */
/* Author: Michael C. Nechyba                                                */
/* Last modified: 6/26/2001                                                  */
/*                                                                           */
/* File: "geometry.h"                                                        */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/

/* Bounding box structure */

typedef struct bounding_box_struct {
  int xmin;
  int xmax;
  int ymin;
  int ymax;
} BOUNDING_BOX;

/*****************************************************************************/

typedef struct trapezoid_struct {
  double c1[3];
  double c2[3];
  double c3[3];
  double c4[3];
} TRAPEZOID;

/*****************************************************************************/

typedef struct rotation_matrix_struct {
  double r11, r12, r13;
  double r21, r22, r23;
  double r31, r32, r33;
} RotationMatrix;

/*****************************************************************************/

typedef struct translation_vector_struct {
  double x, y, z;
} TranslationVector;

/*****************************************************************************/

typedef struct quaternion_struct {
  double s, x, y, z;
} Quaternion;

/*****************************************************************************/

/* convex hull structures */

typedef struct vertix_struct {
  int label;
  double angle;
  DPOINT p;
  struct vertix_struct *next;
  struct vertix_struct *pred;
} VERTIX;

/*****************************************************************************/

extern int rotation_euler(double *vout, double *vin, 
			  double alpha, double beta, double gamma); 
extern int rotation_fixed(double *vout, double *vin, 
			  double alpha, double beta, double gamma);
extern int rotation_euler_trans(double *vout, double *vin, 
				double alpha, double beta, double gamma);
extern int translation(double *vout, double *vin, 
		       double x0, double y0, double z0); 
extern int forward_transformation(double *vout, double *vin, 
				  double alpha, double beta, double gamma, 
				  double x0, double y0, double z0);
extern int inverse_transformation(double *vout, double *vin, 
				  double alpha, double beta, double gamma, 
				  double x0, double y0, double z0);
extern double dot_local(double *v1, double *v2);
extern int cross(double *vout, double *v1, double *v2);
extern int normalize_local(double *vout, double *vin);
extern int plane_equation(double *v, TRAPEZOID *zoid);
extern int rectangles_overlap(BOUNDING_BOX *b1, BOUNDING_BOX *b2);
extern int check_overlap_corner(int x, int y, BOUNDING_BOX *b);
extern int overlap_merge(BOUNDING_BOX *b1, BOUNDING_BOX *b2);

/*****************************************************************************/

extern int copy_rotation_matrix(RotationMatrix *in, RotationMatrix *out);
extern int identity_rotation_matrix(RotationMatrix *m);
extern int rotation_matrix_x(RotationMatrix *m, double alpha);
extern int rotation_matrix_y(RotationMatrix *m, double alpha);
extern int rotation_matrix_z(RotationMatrix *m, double alpha);
extern int display_rotation_matrix(FILE *fp, RotationMatrix *m);
extern int display_rotation_constraints(FILE *fp, RotationMatrix *m);
extern int rXt(TranslationVector *out, RotationMatrix *m, 
	       TranslationVector *in);
extern int rotation_matrix_angle_axis(RotationMatrix *m, 
				      TranslationVector *k, double theta);
extern int multiply_rotation_matrices(RotationMatrix *m,
				      RotationMatrix *a,
				      RotationMatrix *b);
extern int convert_r_to_q(RotationMatrix *m, Quaternion *q);
extern int convert_q_to_r(Quaternion *q, RotationMatrix *m);
extern int normalize_rotation_matrix(RotationMatrix *m);

/*****************************************************************************/

extern double length_of_line(DPOINT a, DPOINT b);
extern double area_of_triangle_from_sides(double a, double b, double c);
extern double area_of_triangle(DPOINT v1, DPOINT v2, DPOINT v3);
extern double area_of_four_sided_polygon(DPOINT v1, DPOINT v2, DPOINT v3, 
					 DPOINT v4);

/*****************************************************************************/

extern int compute_convex_hull(DPOINT *out, DPOINT *p, int n);
extern double compute_convex_angle(DPOINT *centroid, DPOINT *p);
extern int sort_convex_angles(VERTIX *v, int n);
extern int left_hand_turn(VERTIX *v1, VERTIX *v2, VERTIX *v3);
extern double compute_area_of_convex_hull(DPOINT *p, int n);

/*****************************************************************************/

extern int convert_rotation_matrix(RotationMatrix *m, double r[3][3]);
extern int compute_zyx_euler_angles(double *abg, RotationMatrix *m);
