/*****************************************************************************/
/*                                                                           */
/* Author: Michael C. Nechyba                                                */
/* Last modified: 2/13/2000                                                  */
/*                                                                           */
/* File: "median.c"                                                          */
/*                                                                           */
/*****************************************************************************/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>

#ifdef INCLUDE_JAVA
#include "GatorCameraClient.h"
#endif

#include "acticast_include.h"

/*****************************************************************************/

/* quick insertion sort from Numerical Recipes (short lists) */

int my_sort(int *vec, int n)
{
  int i, j, x;

  for(j = 1; j < n; j++) {
    x = vec[j];
    i = j - 1;
    while(i >= 0 && vec[i] > x) {
      vec[i+1] = vec[i];
      i--;
    }
    vec[i+1] = x;
  }

  return(1);
}

/*****************************************************************************/

int median_int(int *vec, int n) 
{
  double x;
  int med;

  my_sort(vec, n);

  /* qsort(vec, n, sizeof(int), compare_int); */

  if(n % 2 == 0) {
    x = ((double) (vec[n/2] + vec[n/2 - 1]))/2.;
    med = (int) (x + .6);
  } else {
    med = vec[n/2];
  }

  return(med);
}

/*****************************************************************************/

int compare_int(int *a, int *b)
{
  if (*a > *b)
    return (1);
  if (*a < *b)
    return (-1);
  return (0);
}

/*****************************************************************************/

int median_filter(PPM *imin, PPM *imout, int n, int t)
{
  int i, j, count = 0;
  
  limit_dimensions(imin);
  copy_dimensions(imin, imout);
  
  for(i = imin->ymin; i < imin->ymax; i++)
    for(j = imin->xmin; j < imin->xmax; j++)
      count += median_filter_one(imin, imout, n, i, j, t);
  
  //fprintf(stderr, "count = %d (%d, %d, %d %d)\n", count,
  //imin->ymin, imin->ymax, imin->xmin, imin->xmax);

  return(count);
}

/*****************************************************************************/

int median_filter_one(PPM *imin, PPM *imout, int n, int i, int j, int t)
{
  int k, l, count, ind;
  int t_count, middle;
  int rmed, gmed, bmed;
  int *r, *g, *b, *rs, *gs, *bs;

  r = &Red[0];
  g = &Green[0];
  b = &Blue[0];

  rs = &RedSort[0];
  gs = &GreenSort[0];
  bs = &BlueSort[0];
  count = 0;

  t_count = 0;
  t *= 3;
  middle = (2*n+1)*(2*n+1)/2 + 1;

  for(k = -n; k <= n; k++) {
    for(l = -n; l <= n; l++) {
      if((i+k) >= imin->ymin && (i+k) < imin->ymax && 
	 (j+l) >= imin->xmin && (j+l) < imin->xmax) {
	r[count] = convert_char(imin->p[j+l][i+k][0]);
	g[count] = convert_char(imin->p[j+l][i+k][1]);
	b[count] = convert_char(imin->p[j+l][i+k][2]);

	/*
	  fprintf(stdout, "(x, y) = (%d, %d): %d + %d + %d = %d\n",
	  j-imin->xmin, i-imin->ymin, r[count], g[count], b[count], 
	  r[count] + g[count] + b[count]);
	*/

	if((r[count] + g[count] + b[count]) < t) {
	  t_count++;
	  if(t_count >= middle) {
	    imout->p[j][i][0] = 0; 
	    imout->p[j][i][1] = 0;
	    imout->p[j][i][2] = 0; /* = 255 --> debugging */
	    
	    return(0);
	  }
	}

	rs[count] = r[count];
	gs[count] = g[count];
	bs[count] = b[count];

	count++;
      }
    }
  }
      
  if(t_count >= count/2+1) {
    imout->p[j][i][0] = 0; 
    imout->p[j][i][1] = 0;
    imout->p[j][i][2] = 0; /* = 255 --> debugging */
    
    return(0);
  }

  rs[count] = r[count];
  rmed = median_int(&rs[0], count);
  gmed = median_int(&gs[0], count);
  bmed = median_int(&bs[0], count);
  
  ind = find_closest(rmed, gmed, bmed, &r[0], &g[0], &b[0], count);

  imout->p[j][i][0] = r[ind];
  imout->p[j][i][1] = g[ind];
  imout->p[j][i][2] = b[ind];

  /* fprintf(stdout, "OUT: %d %d %d\n", r[ind], g[ind], b[ind]); */

  return(1);
}

/*****************************************************************************/

int find_closest(int rmed, int gmed, int bmed,
		 int *rv, int *gv, int *bv, int count)
{
  int i, d, dindex, dmin;

  dmin = 255*255*3;

  /* fprintf(stdout, "d = "); */

  for(i = 0; i < count; i++) {
    d = (rv[i] - rmed)*(rv[i] - rmed) +
      (gv[i] - gmed)*(gv[i] - gmed) +
      (bv[i] - bmed)*(bv[i] - bmed);

    /* fprintf(stdout, "%d ", d); */

    if(d < dmin) {
      dindex = i;
      dmin = d;
    }
  }

  /* fprintf(stdout, "--> %d\n", dindex); */

  return(dindex);
}

/*****************************************************************************/

/* NOT YET OPTIMIZED */

int median_filter_one_raw(RAW *imin, PPM *imout, int n, int i, int j)
{
  char *buf;
  int k, l, count, ind, loc;
  int rmed, gmed, bmed;
  int *r, *g, *b, *rs, *gs, *bs;

  r = &Red[0];
  g = &Green[0];
  b = &Blue[0];

  rs = &RedSort[0];
  gs = &GreenSort[0];
  bs = &BlueSort[0];
  
  count = 0;

  for(k = -n; k <= n; k++) {
    for(l = -n; l <= n; l++) {
      if((i+k) >= imin->ymin && (i+k) < imin->ymax && 
	 (j+l) >= imin->xmin && (j+l) < imin->xmax) {
	loc = ((i+k)*imin->xdim + (j+l))*4;
	buf = &imin->p[loc];

	r[count] = convert_char(buf[2]);
	g[count] = convert_char(buf[1]);
	b[count] = convert_char(buf[0]);

	rs[count] = r[count];
	gs[count] = g[count];
	bs[count] = b[count];

	count++;
      }
    }
  }
      
  rmed = median_int(&rs[0], count);
  gmed = median_int(&gs[0], count);
  bmed = median_int(&bs[0], count);
  
  ind = find_closest(rmed, gmed, bmed, &r[0], &g[0], &b[0], count);

  imout->p[j][i][0] = r[ind];
  imout->p[j][i][1] = g[ind];
  imout->p[j][i][2] = b[ind];

  return(1);
}

/*****************************************************************************/

/* NOT YET OPTIMIZED */

int median_filter_raw(RAW *imin, PPM *imout, int n)
{
  int i, j;
  
  limit_dimensions_raw(imin);
  copy_dimensions_raw_to_ppm(imin, imout);
  
  for(i = imin->ymin; i < imin->ymax; i++)
    for(j = imin->xmin; j < imin->xmax; j++)
      median_filter_one_raw(imin, imout, n, i, j);
  
  return(1);
}

/*****************************************************************************/

int median_filter_ppm2(PPM2 *in, PPM2 *out, int n, int t)
{
  int i, j, j1, count = 0;
  
  limit_dimensions(&in->im);
  copy_mask_ppm2_to_ppm2(in, out);
  
  for(i = 0; i < in->im.ydim; i++) {
    for(j1 = 0; j1 < in->m.line[i].n; j1+=2) {
      for(j = in->m.line[i].c[j1].x; j <= in->m.line[i].c[j1+1].x; j++) {
	count += median_filter_one_ppm2(in, out, n, i, j, t);
      }
    }
  }
  
  return(count);
}

/*****************************************************************************/

int median_filter_red_channel_ppm2(PPM2 *in, PPM2 *out, int n, int t)
{
  int i, j, j1, count = 0;
  
  limit_dimensions(&in->im);
  copy_mask_ppm2_to_ppm2(in, out);
  
  for(i = 0; i < in->im.ydim; i++) {
    for(j1 = 0; j1 < in->m.line[i].n; j1+=2) {
      for(j = in->m.line[i].c[j1].x; j <= in->m.line[i].c[j1+1].x; j++) {
	count += median_filter_red_channel_one_ppm2(in, out, n, i, j, t);
      }
    }
  }
  
  return(count);
}

/*****************************************************************************/

int median_filter_one_ppm2(PPM2 *in, PPM2 *out, int n, int i, int j, int t)
{
  int k, l, count, ind;
  int t_count, middle;
  int rmed, gmed, bmed;
  int *r, *g, *b, *rs, *gs, *bs;

  r = &Red[0];
  g = &Green[0];
  b = &Blue[0];

  rs = &RedSort[0];
  gs = &GreenSort[0];
  bs = &BlueSort[0];
  count = 0;

  t_count = 0;
  t *= 3;
  middle = (2*n+1)*(2*n+1)/2 + 1;

  for(k = -n; k <= n; k++) {
    for(l = -n; l <= n; l++) {
      if((i+k) >= 0 && (i+k) < in->im.ydim && 
	 (j+l) >= 0 && (j+l) < in->im.xdim) {
	r[count] = convert_char(in->im.p[j+l][i+k][0]);
	g[count] = convert_char(in->im.p[j+l][i+k][1]);
	b[count] = convert_char(in->im.p[j+l][i+k][2]);

	/*
	  fprintf(stdout, "(x, y) = (%d, %d): %d + %d + %d = %d\n",
	  j-imin->xmin, i-imin->ymin, r[count], g[count], b[count], 
	  r[count] + g[count] + b[count]);
	*/

	if((r[count] + g[count] + b[count]) < t) {
	  t_count++;
	  if(t_count >= middle) {
	    out->im.p[j][i][0] = 0; 
	    out->im.p[j][i][1] = 0;
	    out->im.p[j][i][2] = 0; /* = 255 --> debugging */
	    
	    return(0);
	  }
	}

	rs[count] = r[count];
	gs[count] = g[count];
	bs[count] = b[count];

	count++;
      }
    }
  }
      
  if(t_count >= count/2+1) {
    out->im.p[j][i][0] = 0; 
    out->im.p[j][i][1] = 0;
    out->im.p[j][i][2] = 0; /* = 255 --> debugging */
    
    return(0);
  }

  rs[count] = r[count];
  rmed = median_int(&rs[0], count);
  gmed = median_int(&gs[0], count);
  bmed = median_int(&bs[0], count);
  
  ind = find_closest(rmed, gmed, bmed, &r[0], &g[0], &b[0], count);

  out->im.p[j][i][0] = r[ind];
  out->im.p[j][i][1] = g[ind];
  out->im.p[j][i][2] = b[ind];

  /* fprintf(stdout, "OUT: %d %d %d\n", r[ind], g[ind], b[ind]); */

  return(1);
}

/*****************************************************************************/

int median_filter_red_channel_one_ppm2(PPM2 *in, PPM2 *out, 
				       int n, int i, int j, int t)
{
  int k, l, count, ind;
  int t_count, middle;
  int rmed, gmed, bmed;
  int *r, *g, *b, *rs, *gs, *bs;

  r = &Red[0];
  g = &Green[0];
  b = &Blue[0];

  rs = &RedSort[0];
  gs = &GreenSort[0];
  bs = &BlueSort[0];
  count = 0;

  t_count = 0;
  t *= 3;
  middle = (2*n+1)*(2*n+1)/2 + 1;

  for(k = -n; k <= n; k++) {
    for(l = -n; l <= n; l++) {
      if((i+k) >= 0 && (i+k) < in->im.ydim && 
	 (j+l) >= 0 && (j+l) < in->im.xdim) {
	r[count] = convert_char(in->im.p[j+l][i+k][0]);
	//	g[count] = convert_char(in->im.p[j+l][i+k][1]);
	//b[count] = convert_char(in->im.p[j+l][i+k][2]);

	/*
	  fprintf(stdout, "(x, y) = (%d, %d): %d + %d + %d = %d\n",
	  j-imin->xmin, i-imin->ymin, r[count], g[count], b[count], 
	  r[count] + g[count] + b[count]);
	*/

	if((r[count] + g[count] + b[count]) < t) {
	  t_count++;
	  if(t_count >= middle) {
	    out->im.p[j][i][0] = 0; 
	      out->im.p[j][i][1] = 0;
	    out->im.p[j][i][2] = 0; /* = 255 --> debugging */
	    
	    return(0);
	  }
	}

	rs[count] = r[count];
	gs[count] = g[count];
	bs[count] = b[count];

	count++;
      }
    }
  }
      
  if(t_count >= count/2+1) {
    out->im.p[j][i][0] = 0; 
       out->im.p[j][i][1] = 0;
    out->im.p[j][i][2] = 0; /* = 255 --> debugging */
    
    return(0);
  }

  rs[count] = r[count];
  rmed = median_int(&rs[0], count);
  gmed = median_int(&gs[0], count);
  bmed = median_int(&bs[0], count);
  
  ind = find_closest(rmed, gmed, bmed, &r[0], &g[0], &b[0], count);

  out->im.p[j][i][0] = r[ind];
  out->im.p[j][i][1] = g[ind];
  out->im.p[j][i][2] = b[ind];

  /* fprintf(stdout, "OUT: %d %d %d\n", r[ind], g[ind], b[ind]); */

  return(1);
}

/*****************************************************************************/

int median_filter_raw2(RAW2 *in, PPM2 *out, int n, int t)
{
  int i, j, j1, count = 0;
  
  limit_dimensions_raw(&in->im);
  copy_mask_raw2_to_ppm2(in, out);
  
  for(i = 0; i < in->im.ydim; i++) {
    for(j1 = 0; j1 < in->m.line[i].n; j1+=2) {
      for(j = in->m.line[i].c[j1].x; j <= in->m.line[i].c[j1+1].x; j++) {
	count += median_filter_one_raw2(in, out, n, i, j, t);
      }
    }
  }
  
  return(count);
}

/*****************************************************************************/

int median_filter_one_raw2(RAW2 *in, PPM2 *out, int n, int i, int j, int t)
{
  int k, l, count, ind, loc, loc2;
  int t_count, middle;
  int rmed, gmed, bmed;
  int *r, *g, *b, *rs, *gs, *bs;

  r = &Red[0];
  g = &Green[0];
  b = &Blue[0];

  rs = &RedSort[0];
  gs = &GreenSort[0];
  bs = &BlueSort[0];
  count = 0;

  t_count = 0;
  t *= 3;
  middle = (2*n+1)*(2*n+1)/2 + 1;
  //loc = (i*in->im.xdim + j)*4;


  for(k = -n; k <= n; k++) {
    for(l = -n; l <= n; l++) {
      if((i+k) >= 0 && (i+k) < in->im.ydim && 
	 (j+l) >= 0 && (j+l) < in->im.xdim) {
	loc2 = ((i+k)*in->im.xdim + (j+l))*4;
	r[count] = convert_char(in->im.p[loc2+2]);
	g[count] = convert_char(in->im.p[loc2+1]);
	b[count] = convert_char(in->im.p[loc2]);

	/*
	  fprintf(stdout, "(x, y) = (%d, %d): %d + %d + %d = %d\n",
	  j-imin->xmin, i-imin->ymin, r[count], g[count], b[count], 
	  r[count] + g[count] + b[count]);
	*/

	if((r[count] + g[count] + b[count]) < t) {
	  t_count++;
	  if(t_count >= middle) {

	    out->im.p[j][i][0] = 0; 
	    out->im.p[j][i][1] = 0;
	    out->im.p[j][i][2] = 0; /* = 255 --> debugging */
	    
	    return(0);
	  }
	}

	rs[count] = r[count];
	gs[count] = g[count];
	bs[count] = b[count];

	count++;
      }
    }
  }
      
  if(t_count >= count/2+1) {
    out->im.p[j][i][0] = 0; 
    out->im.p[j][i][1] = 0;
    out->im.p[j][i][2] = 0; /* = 255 --> debugging */

    return(0);
  }

  rs[count] = r[count];
  rmed = median_int(&rs[0], count);
  gmed = median_int(&gs[0], count);
  bmed = median_int(&bs[0], count);
  
  ind = find_closest(rmed, gmed, bmed, &r[0], &g[0], &b[0], count);

  out->im.p[j][i][0] = r[ind];
  out->im.p[j][i][1] = g[ind];
  out->im.p[j][i][2] = b[ind];

  /* fprintf(stdout, "OUT: %d %d %d\n", r[ind], g[ind], b[ind]); */

  return(1);
}

/*****************************************************************************/
