
/*****************************************************************************/
/*                                                                           */
/* Author: Michael C. Nechyba                                                */
/* Last modified: 03/25/2001                                                 */
/*                                                                           */
/* File: "xdisplay.h"                                                        */
/*                                                                           */
/*****************************************************************************/

#include <X11/Xlib.h>
#include <assert.h>

/*****************************************************************************/

#define FRAME_LABEL_X_OFFSET 5
#define FRAME_LABEL_Y_OFFSET 10
#define FRAME_LABEL_WIDTH 40
#define FRAME_LABEL_HEIGHT 10

#define MAX_XDISPLAY_STEP_FRAMES 10000

/*****************************************************************************/

Window xdisplay_Window;
GC xdisplay_GC;
XImage *xdisplay_XImage;
Display *xdisplay_Display;
char xdisplay_OutputFile[256];
int xdisplay_FrameCounter;
char *xdisplay_background;
char *xdisplay_blank;
int xdisplay_X_DIM;
int xdisplay_Y_DIM;
char *xdisplay_Blowup_Buffer;
int xdisplay_Blowup_Factor;
int xdisplay_FrameShow;
int xdisplay_FrameMin;
int xdisplay_FrameMax;
char *xdisplay_video;
int xdisplay_video_buffer_size;
int xdisplay_video_pointer;
char xdisplay_VideoOutputFile[256];

/*****************************************************************************/

extern int open_xdisplay(char *im);
extern int update_xdisplay(char *im, 
			   BOUNDING_BOX *prev, int n_prev, 
			   BOUNDING_BOX *curr, int n_curr);
extern int full_update_xdisplay(char *im);
extern int copy_bounding_box_to_char(char *vid, char *im, BOUNDING_BOX *b);
extern int copy_char_to_bounding_box(char *vid, char *im, BOUNDING_BOX *b);
extern int compute_size_of_bounding_box(BOUNDING_BOX *b);
extern int write_compressed_video_file(char *file, char *vid, int n_vid);
extern char *read_compressed_video_file(char *file, char *vid, int *alloc);
extern int update_xdisplay_one_box(char *im, BOUNDING_BOX *b);
extern int play_compressed_video(char *vid, int n_vid, char *ref, char *curr,
				 long delay, int decode, int erase);
extern int play_compressed_video_one_frame(char *im0, char *im1, 
					   char *vid, int vid_pointer,
					   BOUNDING_BOX *prev, int *n_prev,
					   int decode, int erase);
extern int blowup_xdisplay_image(char *out, char *in);
extern int put_one_expanded_pixel(char *out, char *buf, int x, int y);
extern char *initialize_video_buffer(char *vid, int size);
extern int initialize_xdisplay_video_buffer(int size);
extern int update_video_buffer(char *vid, int vid_ptr, int size, 
			       char *frame,
			       BOUNDING_BOX *b, int n);
extern int update_xdisplay_video_buffer(char *frame, BOUNDING_BOX *b, int n);
extern int dump_xdisplay_video_buffer(char *file);
extern int interactive_compressed_video(char *vid, int n_vid, char *ref, 
					char *curr, FILE *fp);
extern int handle_xdisplay_interactive_event(XEvent *e, POINT *click_data, 
					     int *n, FILE *fp);
int step_through_compressed_video(char *vid, int n_vid, char *ref, 
				  char *curr, int erase);
extern int handle_xdisplay_step_event(XEvent *e, int *speed, int state);

/*****************************************************************************/

extern int update_xdisplay_video_buffer_half(char *frame);
extern int update_video_buffer_half(char *vid, int vid_ptr, int size, 
				    char *frame);
extern int copy_bounding_box_to_char_half(char *vid, char *im, 
					  BOUNDING_BOX *b);
