/*****************************************************************************/
/*                                                                           */
/* Author: Michael C. Nechyba                                                */
/* Last modified: 06/04/2003                                                 */
/*                                                                           */
/* File: "partition.c"                                                       */
/*                                                                           */
/*****************************************************************************/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#ifdef INCLUDE_JAVA
#include "GatorCameraClient.h"
#endif

#include "acticast_include.h"

/****************************** Subroutines **********************************/

int allocate_class_one(CLASS_ONE *c1)
{
  c1->mu = allocate_vector(c1->mu, 5);
  c1->xy_sigma = allocate_matrix(c1->xy_sigma, 2, 2);
  c1->inv_xy_sigma = allocate_matrix(c1->inv_xy_sigma, 2, 2);

  c1->col_sigma = allocate_matrix(c1->col_sigma, 3, 3);
  c1->inv_col_sigma = allocate_matrix(c1->inv_col_sigma, 3, 3);

  zero_class_one(c1);

  return(1);
}

/*****************************************************************************/

int zero_class_one(CLASS_ONE *c1)
{
  c1->n = 0;
  zero_vector(c1->mu, 5);
  zero_matrix(c1->xy_sigma, 2, 2);
  zero_matrix(c1->inv_xy_sigma, 2, 2);
  zero_matrix(c1->col_sigma, 3, 3);
  zero_matrix(c1->inv_col_sigma, 3, 3);

  return(1);
}

/*****************************************************************************/

int allocate_class(CLASS *c)
{
  int i;

  for(i = 0; i < MAX_CLASS; i++)
    allocate_class_one(&c->class[i]);

  c->n_class = 0;

  for(i = 0; i < 5; i++)
    c->scale[i] = 1.;
  
  return(1);
}

/*****************************************************************************/

int zero_class(CLASS *c)
{
  int i;

  for(i = 0; i < c->n_class; i++)
    zero_class_one(&c->class[i]);

  return(1);
}

/*****************************************************************************/

int initialize_class(CLASS *c, int xgrid, int ygrid, int allocate_flag)
{
  int x, y, count = 0;
  double xdiff, ydiff;
  CLASS_ONE *c1;

  xdiff = 1./((double) xgrid);
  ydiff = 1./((double) ygrid);
  
  if(allocate_flag == 0)
    allocate_class(c);

  for(x = 0; x < xgrid; x++) {
    for(y = 0; y < ygrid; y++) {
      c1 = &c->class[count];
      zero_class_one(c1);
      c1->mu[0] = xdiff*(((double) x)+0.5);
      c1->mu[1] = ydiff*(((double) y)+0.5);
      c1->mu[2] = 0.5;
      c1->mu[3] = 0.5;
      c1->mu[4] = 0.5;
      c1->xy_sigma[0][0] = 1.;
      c1->xy_sigma[1][1] = 1.;
      c1->col_sigma[0][0] = 1.;
      c1->col_sigma[1][1] = 1.;
      c1->col_sigma[2][2] = 1.;
      compute_inverse_sigmas(c1);
      count++;
    }
  }

  c->n_class = count;

  return(count);
}

/*****************************************************************************/

int dump_class_file(char *file, CLASS *c)
{
  int i;
  FILE *fp;
  CLASS_ONE *c1;

  fp = fopen(file, "w");
  
  for(i = 0; i < c->n_class; i++) {
    c1 = &c->class[i];
    
    display_vector(fp, c1->mu, 5, 1, 15);
    display_matrix(fp, c1->xy_sigma, 2, 2, 15);
    display_matrix(fp, c1->col_sigma, 3, 3, 15);
  }

  fclose(fp);

  return(1);
}

/*****************************************************************************/

int read_class_file(char *file, CLASS *c, int allocate_flag)
{
  int count = 0;
  double tmp;
  CLASS_ONE *c1;
  FILE *fp;

  if(allocate_flag == 0)
    allocate_class(c);

  fp = fopen(file, "r");

  while(fscanf(fp, "%lf", &tmp) != EOF) {
    c1 = &c->class[count];
    zero_class_one(c1);
    c1->mu[0] = tmp;

    fscanf(fp, "%lf", &c1->mu[1]);
    fscanf(fp, "%lf", &c1->mu[2]);
    fscanf(fp, "%lf", &c1->mu[3]);
    fscanf(fp, "%lf", &c1->mu[4]);
    fscanf(fp, "%lf", &c1->xy_sigma[0][0]);
    fscanf(fp, "%lf", &c1->xy_sigma[0][1]);
    fscanf(fp, "%lf", &c1->xy_sigma[1][0]);
    fscanf(fp, "%lf", &c1->xy_sigma[1][1]);
    
    fscanf(fp, "%lf", &c1->col_sigma[0][0]);
    fscanf(fp, "%lf", &c1->col_sigma[0][1]);
    fscanf(fp, "%lf", &c1->col_sigma[0][2]);
    fscanf(fp, "%lf", &c1->col_sigma[1][0]);
    fscanf(fp, "%lf", &c1->col_sigma[1][1]);
    fscanf(fp, "%lf", &c1->col_sigma[1][2]);
    fscanf(fp, "%lf", &c1->col_sigma[2][0]);
    fscanf(fp, "%lf", &c1->col_sigma[2][1]);
    fscanf(fp, "%lf", &c1->col_sigma[2][2]);
    
    compute_inverse_sigmas(c1);
    count++;
  }
  
  fclose(fp);

  c->n_class = count;
  
  return(count);
}

/*****************************************************************************/

int allocate_ppm_data(PPM_DATA *m, int xdim, int ydim)
{
  m->n = xdim*ydim;
  m->data = allocate_matrix(m->data, m->n, 5);
  m->class = int_allocate_vector(m->class, m->n);

  return(1);
}
/*****************************************************************************/

int convert_ppm_to_data(PPM *im, PPM_DATA *m, int allocate_flag)
{
  int i, x, y, count = 0;
  double xdim1, ydim1;

  if(allocate_flag == 0) {
    allocate_ppm_data(m, im->xdim, im->ydim);
  }

  xdim1 = (double) im->xdim - 1.;
  ydim1 = (double) im->ydim - 1.;

  for(y = 0; y < im->ydim; y++) {
    for(x = 0; x < im->xdim; x++) {
      m->data[count][0] = ((double) x)/xdim1;
      m->data[count][1] = ((double) y)/ydim1;
      m->data[count][2] = (double) convert_char(im->p[x][y][RED]);
      m->data[count][3] = (double) convert_char(im->p[x][y][GREEN]);
      m->data[count][4] = (double) convert_char(im->p[x][y][BLUE]);
      count++;
    }
  }

  for(i = 0; i < m->n; i++) {
    m->data[i][2] = m->data[i][2]/255.;
    m->data[i][3] = m->data[i][3]/255.;
    m->data[i][4] = m->data[i][4]/255.;
  }
  
  return(1);
}

/*****************************************************************************/

int compute_inverse_sigmas(CLASS_ONE *c1)
{
  inverse(c1->inv_xy_sigma,  c1->xy_sigma,  2);
  inverse(c1->inv_col_sigma, c1->col_sigma, 3);

  return(1);
}

/*****************************************************************************/

double weighted_distance(double *v, CLASS_ONE *c1, double *scale)
{
  int i;
  double dist = 0.;
  double vd[5];
  double p00, p01, p11;
  double c00, c01, c02, c11, c12, c22;
  
  p00 = c1->inv_xy_sigma[0][0];
  p01 = c1->inv_xy_sigma[0][1];
  p11 = c1->inv_xy_sigma[1][1];
  
  c00 = c1->inv_col_sigma[0][0];
  c01 = c1->inv_col_sigma[0][1];
  c02 = c1->inv_col_sigma[0][2];
  c11 = c1->inv_col_sigma[1][1];
  c12 = c1->inv_col_sigma[1][2];
  c22 = c1->inv_col_sigma[2][2];

  for(i = 0; i < 5; i++) {
    vd[i] = (v[i] - c1->mu[i])*sqrt(scale[i]);
    //dist += vd[i]*vd[i];
  }

  dist = p00*vd[0]*vd[0] + 2.*p01*vd[0]*vd[1] + p11*vd[1]*vd[1] + 
    c00*vd[2]*vd[2] + 2.*c01*vd[2]*vd[3] + c11*vd[3]*vd[3] + 
    2.*c02*vd[2]*vd[4] + 2.*c12*vd[3]*vd[4] + c22*vd[4]*vd[4];

  return(dist);
}

/*****************************************************************************/

int ppm_classify_one(double *v, CLASS *c, double *dist_out)
{
  int i, min_index;
  double dist, min_dist = 1000000.;
  
  for(i = 0; i < c->n_class; i++) {
    dist = weighted_distance(v, &c->class[i], &c->scale[0]);
    if(dist < min_dist) {
      min_dist = dist;
      min_index = i;
    }
  }
  
  *dist_out = min_dist;
  return(min_index);
}

/*****************************************************************************/

double ppm_classify_all(PPM_DATA *m, CLASS *c)
{
  int i;
  double dist, dist_total = 0.;

  //for(i = 0; i < c->n_class; i++) {
  //identity_matrix(c->class[i].inv_xy_sigma, 2);
  //identity_matrix(c->class[i].inv_col_sigma, 3);
  //}

  for(i = 0; i < m->n; i++) {
    m->class[i] = ppm_classify_one(&m->data[i][0], c, &dist);
    dist_total += dist;
  }

  return(dist_total);
}

/*****************************************************************************/

int ppm_update_class_one(double *v, CLASS_ONE *c1)
{
  c1->n++;

  c1->mu[0] += v[0];
  c1->mu[1] += v[1];
  c1->mu[2] += v[2];
  c1->mu[3] += v[3];
  c1->mu[4] += v[4];

  c1->xy_sigma[0][0] += v[0]*v[0];
  c1->xy_sigma[0][1] += v[0]*v[1];
  c1->xy_sigma[1][0] = c1->xy_sigma[0][1];
  c1->xy_sigma[1][1] += v[1]*v[1];
  
  c1->col_sigma[0][0] += v[2]*v[2];
  c1->col_sigma[0][1] += v[2]*v[3];
  c1->col_sigma[0][2] += v[2]*v[4];
  
  c1->col_sigma[1][0] = c1->col_sigma[0][1];
  c1->col_sigma[1][1] += v[3]*v[3];
  c1->col_sigma[1][2] += v[3]*v[4];
  
  c1->col_sigma[2][0] = c1->col_sigma[0][2];
  c1->col_sigma[2][1] = c1->col_sigma[1][2];
  c1->col_sigma[2][2] += v[4]*v[4];

  return(1);
}

/*****************************************************************************/

int ppm_update_class_all(PPM_DATA *m, CLASS *c)
{
  int i;
  
  zero_class(c);

  for(i = 0; i < m->n; i++)
    ppm_update_class_one(&m->data[i][0], &c->class[m->class[i]]);

  ppm_adjust_class_all(c);

  return(1);
}

/*****************************************************************************/

int ppm_adjust_class_one(CLASS_ONE *c1)
{
  double nd, xd, yd, rd, gd, bd;

  nd = ((double) c1->n);

  c1->mu[0] /= nd;
  c1->mu[1] /= nd;
  c1->mu[2] /= nd;
  c1->mu[3] /= nd;
  c1->mu[4] /= nd;

  xd = c1->mu[0];
  yd = c1->mu[1];
  rd = c1->mu[2];
  gd = c1->mu[3];
  bd = c1->mu[4];
  
  c1->xy_sigma[0][0] = nd/(nd - 1.)*(c1->xy_sigma[0][0]/nd - xd*xd); 
  c1->xy_sigma[0][1] = nd/(nd - 1.)*(c1->xy_sigma[0][1]/nd - xd*yd);
  c1->xy_sigma[1][0] = c1->xy_sigma[0][1];
  c1->xy_sigma[1][1] = nd/(nd - 1.)*(c1->xy_sigma[1][1]/nd - yd*yd);
  
  c1->col_sigma[0][0] = nd/(nd - 1.)*(c1->col_sigma[0][0]/nd - rd*rd); 
  c1->col_sigma[0][1] = nd/(nd - 1.)*(c1->col_sigma[0][1]/nd - rd*gd);
  c1->col_sigma[0][2] = nd/(nd - 1.)*(c1->col_sigma[0][2]/nd - rd*bd);
  
  c1->col_sigma[1][0] = c1->col_sigma[0][1];
  c1->col_sigma[1][1] = nd/(nd - 1.)*(c1->col_sigma[1][1]/nd - gd*gd);
  c1->col_sigma[1][2] = nd/(nd - 1.)*(c1->col_sigma[1][2]/nd - gd*bd);
  
  c1->col_sigma[2][0] = c1->col_sigma[0][2];
  c1->col_sigma[2][1] = c1->col_sigma[1][2];
  c1->col_sigma[2][2] = nd/(nd - 1.)*(c1->col_sigma[2][2]/nd - bd*bd);
  
  c1->xy_sigma[0][0] += PARTITION_EPSILON;
  c1->xy_sigma[1][1] += PARTITION_EPSILON;
  
  c1->col_sigma[0][0] += PARTITION_EPSILON;
  c1->col_sigma[1][1] += PARTITION_EPSILON;
  c1->col_sigma[2][2] += PARTITION_EPSILON;
  
  compute_inverse_sigmas(c1);

  return(1);
}
     
/*****************************************************************************/

int ppm_adjust_class_all(CLASS *c)
{
  int i;
  
  for(i = 0; i < c->n_class; i++)
    ppm_adjust_class_one(&c->class[i]);
  
  return(1);
}

/*****************************************************************************/

int draw_class_picture(RAW *im, int **colors, int n_colors, int *class, int n)
{
  int i, ptr = 0;

  for(i = 0; i < n; i++) {
    im->p[ptr+BLUE_RAW] = colors[class[i] % n_colors][0];
    im->p[ptr+GREEN_RAW] = colors[class[i] % n_colors][1];
    im->p[ptr+RED_RAW] = colors[class[i] % n_colors][2];
    ptr += 4;
  }

  return(1);
}

/*****************************************************************************/

int draw_class_real_colors(RAW *im, CLASS *c, int *class, int n)
{
  int i, ptr = 0;
  int red, green, blue;

  for(i = 0; i < n; i++) {
    red   = limit_int((int) (c->class[class[i]].mu[2]*255. + 0.5), 0, 255);
    green = limit_int((int) (c->class[class[i]].mu[3]*255. + 0.5), 0, 255);
    blue  = limit_int((int) (c->class[class[i]].mu[4]*255. + 0.5), 0, 255);
    im->p[ptr+BLUE_RAW]   = blue;
    im->p[ptr+GREEN_RAW] = green;
    im->p[ptr+RED_RAW] = red;
    ptr += 4;
  }

  return(1);
}

/*****************************************************************************/

int display_class(FILE *fp, CLASS *c)
{
  int i;

  for(i = 0; i < c->n_class; i++) {
    fprintf(fp, ">>>>> %d (%d):\n", i, c->class[i].n);
    display_vector(fp, &c->class[i].mu[0], 5, 0, 5);
    display_matrix(fp, c->class[i].xy_sigma, 2, 2, 5);
    display_matrix(fp, c->class[i].col_sigma, 3, 3, 5);
  }

  return(1);
}

/*****************************************************************************/

int threshold_color_raw(RAW *in, PPM *out, int r, int g, int b,
			int foreground, int background)
{
  int i, j, loc;
  int rtst, gtst, btst;
  char *bufin, *bufout, val;
  
  limit_dimensions_raw(in);
  copy_dimensions_raw_to_ppm(in, out);
  
  for(i = 0; i < in->ydim; i++) {
    for(j = 0; j < in->xdim; j++) {
      loc = (i*in->xdim + j)*4;
      bufin  = &in->p[loc];
      bufout = &out->p[j][i][0];

      rtst = convert_char(bufin[RED_RAW]);
      gtst = convert_char(bufin[GREEN_RAW]);
      btst = convert_char(bufin[BLUE_RAW]);
      
      if(r == rtst && g == gtst && b == btst) {
	val = (char) foreground;
      } else {
	val = (char) background;
      }
      bufout[2] = val;
      bufout[1] = val;
      bufout[0] = val;
    }
  }

  return(1);
}

/*****************************************************************************/

int color_overlay_raw(RAW *in, PPM *t, PPM *out, int r, int g, int b)
{
  int i, j, loc;
  int rtst, gtst, btst;
  char *bufin, *bufout, *buft, val;
  
  
  limit_dimensions_raw(in);
  copy_dimensions_raw_to_ppm(in, out);
  
  for(i = 0; i < in->ydim; i++) {
    for(j = 0; j < in->xdim; j++) {
      loc = (i*in->xdim + j)*4;
      bufin  = &in->p[loc];
      buft  =  &t->p[j][i][0];
      bufout = &out->p[j][i][0];
      
      rtst = convert_char(buft[RED]);
      gtst = convert_char(buft[GREEN]);
      btst = convert_char(buft[BLUE]);
      
      if(r == rtst && g == gtst && b == btst) {
	bufout[RED] = bufin[RED_RAW];
	bufout[GREEN] = bufin[GREEN_RAW];
	bufout[BLUE] = bufin[BLUE_RAW];
      } else {
	bufout[RED] = (char) BACKGROUND_VALUE;
	bufout[GREEN] = (char) BACKGROUND_VALUE;
	bufout[BLUE] = (char) BACKGROUND_VALUE;
      }
    }
  }

  return(1);
}

/*****************************************************************************/

int set_global_variables(GLOBAL *g, char *parameter_file,
			 PARAMETERS *table, int n_params)
{
  int i;

  file_no_exist(parameter_file, "parameter");
  get_parameters(parameter_file, table, n_params);
  file_no_exist(&g->image_file[0], "input image");

  read_ppm_file2(&g->image_file[0], &g->im_ppm.im);
  g->box_all.xmin = 0;
  g->box_all.xmax = g->im_ppm.im.xdim;
  g->box_all.ymin = 0;
  g->box_all.ymax = g->im_ppm.im.ydim;


  if(g->rgb2hsv_flag > 0) {
    convert_rgb_to_hsv(&g->im_ppm.im, &g->hsv_ppm.im);
    ppm2raw(&g->hsv_ppm.im, &g->hsv.im, 0);
    g->curr_ppm = &g->hsv_ppm;
    g->curr = &g->hsv;
    //show_and_save_image(g, &g->hsv.im, 0);
  } else {
    g->curr_ppm = &g->im_ppm;
    g->curr = &g->im;
  }
  g->initial = g->curr;

  ppm2raw(&g->im_ppm.im, &g->im.im, 0);
  //fprintf(stderr, "%d %d %d %d %d %d %d\n",
  //  convert_char(g->im_ppm.im.p[0][0][0]),
  //  convert_char(g->im_ppm.im.p[0][0][1]),
  //  convert_char(g->im_ppm.im.p[0][0][2]),
  //  convert_char(g->im.im.p[0]),
  //  convert_char(g->im.im.p[1]),
  //  convert_char(g->im.im.p[2]),
  //  convert_char(g->im.im.p[3]));


  g->total.im.xdim = g->im_ppm.im.xdim; 
  g->total.im.ydim = g->im_ppm.im.ydim;

  make_single_color_image(&g->total.im, 
			  BACKGROUND_VALUE, 
			  BACKGROUND_VALUE, 
			  BACKGROUND_VALUE);
  set_dimensions(&g->total.im, 0, g->total.im.xdim, 0, g->total.im.ydim);
  ppm2raw(&g->total.im, &g->proc.im, 0);
  ppm2raw(&g->total.im, &g->display1.im, 0);
  ppm2raw(&g->total.im, &g->display2.im, 0);

  if(file_exist(&g->partitioned_file[0])) {
    read_ppm_file2(&g->partitioned_file[0], &g->proc_ppm.im);
    ppm2raw(&g->proc_ppm.im, &g->proc.im, 1);
    g->initial = &g->proc;
  } else {
    if(file_exist(&g->vq_file[0])) {
      read_class_file(&g->vq_file[0], &g->c, 0);
    } else {
      initialize_class(&g->c, g->grid.x, g->grid.y, 0);
    }
    g->colors = int_allocate_matrix(g->colors, MAX_CLASS, 3);
    g->n_colors = g->c.n_class;
    for(i = 0; i < g->n_colors; i++) {
      g->colors[i][0] = random_integer2(0, 255);
      g->colors[i][1] = random_integer2(0, 255);
      g->colors[i][2] = random_integer2(0, 255);
      //fprintf(stderr, "colors = %d %d %d\n", 
      //g->colors[i][0], g->colors[i][1], g->colors[i][2]);
    }
  }

  convert_ppm_to_data(&g->curr_ppm->im, &g->m, 0);
  g->uniq_colors = int_allocate_matrix(g->uniq_colors, MAX_CLASS, 3);
  allocate_blob2(&g->all[0], g->im_ppm.im.xdim*g->im_ppm.im.ydim);
  initialize_blob2(&g->all[0]);
  for(i = 0; i < MAX_SPLIT_BLOBS; i++)
    allocate_blob2(&g->split[i], MAX_SPLIT_BLOB_SIZE);
  allocate_class(&g->clean);
    

  return(1);
}

/*****************************************************************************/

int default_global_variables(GLOBAL *g) 
{
  g->grid.x = 8;
  g->grid.y = 8;
  g->rgb2hsv_flag = 1;
  g->interrupt_flag = 0;
  g->dump_all_flag = -1;
  g->image_counter = 1000;
  g->append_file_flag = -1;
  g->cohesion_scale = 1.;
  sprintf(&g->image_out_prefix[0], "out");
  sprintf(&g->partitioned_file[0], "%c", 0);
  sprintf(&g->vq_file[0], "%c", 0);
  sprintf(&g->output_data_file[0], "data.txt");

  return(1);
}

/*****************************************************************************/

int partition_image_global(GLOBAL *g)
{
  partition_image_global_with_class(g, &g->c, g->colors, g->n_colors);
  return(1);
}

/*****************************************************************************/

int partition_image_global_with_class(GLOBAL *g, CLASS *c,
				      int **colors, int n_colors)
{
  int done = 0, count = 0;
  double err_prev, err_curr, err_eps = 0.00001, derr;

  //print_partition_instructions(stdout);

  err_prev = 1000000000.;

  while(!done) {
    count++;
    err_curr = ppm_classify_all(&g->m, c);
    fprintf(stdout, "Iteration #%d: %.10f\r", count, err_curr);
    fflush(stdout);

    draw_class_real_colors(&g->proc.im, c, &g->m.class[0], g->m.n);

    draw_class_picture(&g->proc.im, colors, n_colors, 
		       &g->m.class[0], g->m.n);
    show_and_save_image(g, &g->proc.im, 0);

    derr = fabs(err_curr - err_prev)/err_curr;
    if(derr < err_eps || g->interrupt_flag == 1)
      done = 1;
    err_prev = err_curr;

    ppm_update_class_all(&g->m, c);
    //sleep(5);
  }

  return(1);
}

/*****************************************************************************/

int show_and_save_image(GLOBAL *g, RAW *im, int force)
{
  char str[256];

  full_update_xdisplay(im->p);

  if(g->dump_all_flag > 0 || force > 0) {
    sprintf(&str[0], "%s%d.ppm", &g->image_out_prefix[0], g->image_counter);
    write_ppm_file_from_raw(&str[0], im->p, 
			    g->im_ppm.im.xdim, g->im_ppm.im.ydim);
    g->image_counter++;
  }
  
  return(1);
}
  
/*****************************************************************************/

int print_partition_instructions(FILE *fp)
{
  fprintf(fp, "\n\nInstructions:\n\n");
  fprintf(fp, "When the image looks sufficiently partitioned,\n");
  fprintf(fp, "press Control-C, or wait for the program to decide\n");
  fprintf(fp, "when to stop.\n");
  fprintf(fp, "\n\n");

  return(1);
}

/*****************************************************************************/

int extract_image_global(GLOBAL *g)
{
  int done = 0;
  FILE *fp;

  print_extraction_instructions(stdout);
  full_update_xdisplay(g->proc.im.p);

  g->event_state = PROCESSED_IMAGE_FLAG;
  g->interrupt_flag = 0;
  XSelectInput(xdisplay_Display, xdisplay_Window, ButtonPress);

  while(!done) {
    XWindowEvent(xdisplay_Display, xdisplay_Window, ButtonPress, 
		 &g->event);
    done = handle_extract_event(g);
  }

  if(g->all[0].n <= 0) {
    fprintf(stdout, "No data extracted.\n");
  } else {
    if(g->append_file_flag > 0) {
      fp = fopen(&g->output_data_file[0], "a");
      fprintf(stdout, "\n\nAppending data output file (%s).\n",
	      &g->output_data_file[0]);
    } else {
      fp = fopen(&g->output_data_file[0], "w");
      fprintf(stdout, "\n\nWriting data output file (%s).\n",
	      &g->output_data_file[0]);
    }
    
    display_blob2_local(fp, &g->all[0]);
    fclose(fp);
    if(g->dump_all_flag <= 0) {
      show_and_save_image(g, &g->display2.im, 1);
    }
  }

  return(1);
}

/*****************************************************************************/

int print_extraction_instructions(FILE *fp)
{
  fprintf(fp, "\n\nInstructions:\n\n");
  fprintf(fp, "When the partitioned image is displayed, use Button 1\n");
  fprintf(fp, "to select a specific region (class). This will bring up\n");
  fprintf(fp, "a thresholded image which displays only the selected class.\n");
  fprintf(fp, "\n");
  fprintf(fp, "Whent the thresholded image is displayed, use Button 2 to \n");
  fprintf(fp, "select sub regions to be included in the color model data,\n");
  fprintf(fp, "and use Button 1 to return back to the partitioned image.\n");
  fprintf(fp, "\n");
  fprintf(fp, "When you are done, press Button 3, and the selected data\n");
  fprintf(fp, "will be dumped to the data output file.\n");
  fprintf(fp, "\n\n");

  return(1);
}

/*****************************************************************************/

int handle_extract_event(GLOBAL *g)
{
  XEvent *e; 
  int x, y, button;

  e = &g->event;
  x = e->xbutton.x;
  y = e->xbutton.y;
  button = e->xbutton.button;

  if(button == Button3) {
    if(g->event_state == THRESHOLD_IMAGE_FLAG) {
      clean_up_mask_image_partition(g);
    }
    return(1);
  }

  if(button == Button1 && g->event_state == PROCESSED_IMAGE_FLAG) {
    generate_mask_image_partition(g, x, y);
  } else if(button == Button2 && g->event_state == THRESHOLD_IMAGE_FLAG) {
    gather_pixel_information_partition(g, x, y);
  } else if (e->xbutton.button == Button1) {
    clean_up_mask_image_partition(g);
    full_update_xdisplay(g->proc.im.p);
  }

  return(0);
}

/*****************************************************************************/

int merge_specialized(PPM *im1, PPM *im2, PPM *out)
{
  int i, j;
  char *buf1, *buf2, *bout;

  limit_dimensions(im1);
  limit_dimensions(im2);
  copy_dimensions(im1, out);
  
  for(i = 0; i < im1->ydim; i++) {
    for(j = 0; j < im1->xdim; j++) {
      buf1 = &im1->p[j][i][0];
      buf2 = &im2->p[j][i][0];
      bout = &out->p[j][i][0];

      if(!(buf1[0] == (char) BACKGROUND_VALUE &&
	   buf1[1] == (char) BACKGROUND_VALUE &&
	   buf1[2] == (char) BACKGROUND_VALUE)) {
	bout[0] = buf1[0];
	bout[1] = buf1[1];
	bout[2] = buf1[2];
      } else {
	bout[0] = buf2[0];
	bout[1] = buf2[1];
	bout[2] = buf2[2];
      }
    }
  }

  return(1);
}

/*****************************************************************************/

int eliminate_unselected_foreground_pixels(PPM *im)
{
  int i, j;
  char *buf;

  limit_dimensions(im);
  
  for(i = 0; i < im->ydim; i++) {
    for(j = 0; j < im->xdim; j++) {
      buf = &im->p[j][i][0];

      if(buf[0] == (char) FOREGROUND_VALUE &&
	 buf[1] == (char) FOREGROUND_VALUE &&
	 buf[2] == (char) FOREGROUND_VALUE) {
	buf[0] = (char) BACKGROUND_VALUE;
	buf[1] = (char) BACKGROUND_VALUE;
	buf[2] = (char) BACKGROUND_VALUE;
      }
    }
  }

  return(1);
}

/*****************************************************************************/

int display_blob2_local(FILE *fp, BLOB2 *b)
{
  int i;

  for(i = 0; i < b->n; i++)
    fprintf(fp, "%d %d %d %d %d\n",
            (int) b->rgb_t[RED][i],
            (int) b->rgb_t[GREEN][i],
            (int) b->rgb_t[BLUE][i],
            (int) b->xy_t[0][i],
            (int) b->xy_t[1][i]);

  return(1);
}

/*****************************************************************************/

int clean_up_mask_image_partition(GLOBAL *g)
{
  eliminate_unselected_foreground_pixels(&g->total.im);

  ppm2raw(&g->total.im, &g->display2.im, 1);
  show_and_save_image(g, &g->display2.im, 0);

  sleep(1);

  g->event_state = PROCESSED_IMAGE_FLAG;
  
  return(1);
}

/*****************************************************************************/

int gather_pixel_information_partition(GLOBAL *g, int x, int y)
{
  int red, green, blue, prev_size;
  int label = 1;

  prev_size = g->all[0].n;

  red =   convert_char(g->total.im.p[x][y][RED]);
  green = convert_char(g->total.im.p[x][y][GREEN]);
  blue =  convert_char(g->total.im.p[x][y][BLUE]);

  if(red == FOREGROUND_VALUE && 
     green == FOREGROUND_VALUE && 
     blue == FOREGROUND_VALUE) {
    collect_statistics_for_single_blob2(g->curr, &g->tmp2, x, y, 
					label, FOREGROUND_VALUE, &g->all[0]);
    color_overlay_raw(&g->im.im, &g->tmp2.im, &g->tmp3.im, 
		      label, FOREGROUND_VALUE, FOREGROUND_VALUE);
    merge_specialized(&g->tmp3.im, &g->total.im, &g->total.im);
    ppm2raw(&g->total.im, &g->display2.im, 1);
    show_and_save_image(g, &g->display2.im, 0);
    fprintf(stdout, "Added blob of %d pixels. Total size = %d pixels.\n", 
	    g->all[0].n - prev_size, g->all[0].n);
  }

  return(1);
}

/*****************************************************************************/

int generate_mask_image_partition(GLOBAL *g, int x, int y) 
{
  int loc, red, green, blue;
  
  loc = (y*g->proc.im.xdim + x)*4;

  red =   convert_char(g->proc.im.p[loc+RED_RAW]);
  green = convert_char(g->proc.im.p[loc+GREEN_RAW]);
  blue =  convert_char(g->proc.im.p[loc+BLUE_RAW]);

  threshold_color_raw(&g->proc.im, &g->tmp1.im, red, green, blue,
		      FOREGROUND_VALUE, BACKGROUND_VALUE);
  create_mask(&g->tmp1.m, &g->box_all, 1);
  median_filter_ppm2(&g->tmp1, &g->tmp2, 1, 128);
  merge_specialized(&g->total.im, &g->tmp2.im, &g->total.im);
  /*
    threshold_color_raw(&g->proc.im, &g->tmp2.im, red, green, blue,
    FOREGROUND_VALUE, BACKGROUND_VALUE);
    create_mask(&g->tmp2.m, &g->box_all, 1);
    merge_specialized(&g->total.im, &g->tmp2.im, &g->total.im);
  */
  ppm2raw(&g->total.im, &g->display1.im, 1);
  show_and_save_image(g, &g->display1.im, 0);
  g->event_state = THRESHOLD_IMAGE_FLAG;

  return(1);
}

/*****************************************************************************/

int extract_uniq_colors(RAW *im, int **colors)
{
  int i, j, k, loc, n, red, green, blue, flag;
  char *buf;

  n = 0;

  for(i = 0; i < im->ydim; i++) {
    for(j = 0; j < im->xdim; j++) {
      loc = (i*im->xdim + j)*4;
      buf = &im->p[loc];

      red   = convert_char(buf[RED_RAW]);
      green = convert_char(buf[GREEN_RAW]);
      blue  = convert_char(buf[BLUE_RAW]);

      if(n == 0) {
	colors[n][0] = red;
	colors[n][1] = green;
	colors[n][2] = blue;
	n++;
      } else {
	flag = 0;
	k = n-1;
	while(flag == 0) {
	  if(red == colors[k][0] && 
	     green == colors[k][1] &&
	     blue == colors[k][2]) {
	    flag = 1;
	  }
	  
	  if(k == 0 && flag == 0)
	    flag = 2;
	  else
	    k--;
	}

	if(flag == 2) {
	  colors[n][0] = red;
	  colors[n][1] = green;
	  colors[n][2] = blue;
	  n++;
	}
      }
    }
  }

  return(n);
}

/*****************************************************************************/

int clean_up_partitioned_image(GLOBAL *g)
{
  int i; 

  fprintf(stdout, "\nSplitting into contiguous classes.\n");

  g->clean.n_class = 0;
  g->n_uniq = extract_uniq_colors(&g->proc.im, g->uniq_colors);
  
  //fprintf(stdout, "Number pf initial classes: %d.\n", g->n_uniq);
  for(i = 0; i < g->n_uniq; i++) {
    clean_up_one_partition(g, &g->uniq_colors[i][0]);
    fprintf(stdout, "Total classes = %d.\r", g->clean.n_class);
    fflush(stdout);
  }

  for(i = g->n_uniq; i < g->clean.n_class; i++) {
      g->uniq_colors[i][0] = random_integer2(0, 255);
      g->uniq_colors[i][1] = random_integer2(0, 255);
      g->uniq_colors[i][2] = random_integer2(0, 255);
  }
  g->n_uniq = g->clean.n_class;

  g->clean.scale[2] = 1./g->cohesion_scale;
  g->clean.scale[3] = 1./g->cohesion_scale;
  g->clean.scale[4] = 1./g->cohesion_scale;

  ppm_classify_all(&g->m, &g->clean);
  draw_class_picture(&g->proc.im, g->uniq_colors, g->n_uniq, 
		     &g->m.class[0], g->m.n);
  
   return(g->clean.n_class);
}

/*****************************************************************************/

int clean_up_one_partition(GLOBAL *g, int *color)
{
  int i, n;

  threshold_color_raw(&g->proc.im, &g->tmp1.im, 
		      color[0], color[1], color[2], 255, 0);
  create_mask(&g->tmp1.m, &g->box_all, 1);
  median_filter_ppm2(&g->tmp1, &g->tmp2, 1, 128);
  n = detect_blobs_ppm2(g->curr, &g->tmp2, &g->split[0], 
			MIN_PARTITION_BLOB_SIZE);
  ppm2raw(&g->tmp2.im, &g->display1.im, 1);
  show_and_save_image(g, &g->display1.im, 0);

  for(i = 0; i < n; i++) {
    compute_class_from_blob2(&g->clean, &g->split[i], 
			     g->proc.im.xdim, g->proc.im.ydim);
  }

  return(1);
}

/*****************************************************************************/

int compute_class_from_blob2(CLASS *c, BLOB2 *b, int xdim, int ydim)
{
  int i;
  CLASS_ONE *c1;
  double v[5], xdim1, ydim1;

  c1 = &c->class[c->n_class];
  zero_class_one(c1);

  xdim1 = (double) (xdim - 1);
  ydim1 = (double) (ydim - 1);

  for(i = 0; i < b->n; i++) {
    v[0] = b->xy_t[0][i]/xdim1;
    v[1] = b->xy_t[1][i]/ydim1;
    v[2] = b->rgb_t[0][i]/255.;
    v[3] = b->rgb_t[1][i]/255.;
    v[4] = b->rgb_t[2][i]/255.;
    ppm_update_class_one(&v[0], c1);
  }

  ppm_adjust_class_one(c1);
  c->n_class++;

  return(1);
}

/*****************************************************************************/

int collect_statistics_for_single_blob2(RAW2 *frame, PPM2 *t, 
					int x, int y, int label, int tst,
					BLOB2 *b)
{
  int i, j;
  int tmp_x, tmp_y, tmp_n;
  int status, n_stack, flag; //  recursive call limit flags
  int flag2;

  b[0].first.x = x;
  b[0].first.y = y;
  b[0].label = label;

  flag = 0;
  n_stack = 0;
  status = 1;
  tmp_x = x; tmp_y = y;

  while(flag == 0) {
    new_collect_blob_data_ppm2(frame, t, tmp_x, tmp_y, 
			       label, tst, &b[0], 
			       &status, &n_stack);
    flag2 = 0;
    for(i = 0; i < t->im.ydim; i++)
      for(j = 0; j < t->im.xdim; j++) {
	if(t->im.p[j][i][0] == tst) {
	  if(check_for_completion(&t->im, j, i, label) == 1) {
	    flag2 = 1;
	    tmp_x = j; 
	    tmp_y = i; 
	    j = t->im.xdim;
	    i = t->im.ydim;
	  }
	}
      }
    if(flag2 == 0)
      flag = 1;
    else {
      n_stack = 0;
      status = 1;
    }
  }
  
  //fprintf(stderr, ">>b[0].n = %d\n", b[0].n);
  return(b[0].n);
}

/*****************************************************************************/

int check_for_completion(PPM *t, int x, int y, int label)
{
  int tst;

  tst = label;

#ifdef EIGHT_CONNECTED  
  if((x-1) >= 0 && (y-1) >= 0) 
    if(t->p[x-1][y-1][0] == (char) tst)
      return(1);
#endif
  
  if((y-1) >= 0)
    if(t->p[x][y-1][0] == (char) tst) 
      return(1);
  
#ifdef EIGHT_CONNECTED
  if((x+1) < t->xdim && (y-1) >= 0) 
    if(t->p[x+1][y-1][0] == (char) tst) 
      return(1);
#endif
  
  if((x-1) >= 0) 
    if(t->p[x-1][y][0] == (char) tst) 
      return(1);
  
  if((x+1) < t->xdim) 
    if(t->p[x+1][y][0] == (char) tst) 
      return(1);
  
#ifdef EIGHT_CONNECTED
  if((x-1) >= 0 && (y+1) < t->ydim) 
    if(t->p[x-1][y+1][0] == (char) tst) 
      return(1);
#endif
  
  if((y+1) < t->ydim) 
    if(t->p[x][y+1][0] == (char) tst) 
      return(1);
  
#ifdef EIGHT_CONNECTED
  if((x+1) < t->xdim && (y+1) < t->ydim)
    if(t->p[x+1][y+1][0] == (char) tst) 
      return(1);
#endif
  
  return(0);
}

/*****************************************************************************/
