/*****************************************************************************/
/*                                                                           */
/* Author: Michael C. Nechyba                                                */
/* Last modified: 04/09/2001                                                 */
/*                                                                           */
/* File: "partition.h"                                                       */
/*                                                                           */
/*****************************************************************************/

#define MAX_CLASS 1024
#define MAX_SPLIT_BLOBS 20
#define MAX_SPLIT_BLOB_SIZE 300000
#define MIN_PARTITION_BLOB_SIZE 200

#define PARTITION_EPSILON 0.2

#define PROCESSED_IMAGE_FLAG 1
#define THRESHOLD_IMAGE_FLAG 2

#define BACKGROUND_VALUE 225
#define FOREGROUND_VALUE 0

/*****************************************************************************/

typedef struct class_one_struct {
  int n;
  double *mu;             // 5x1 (x,y,[r,h], [g,s], [b,v])
  double **xy_sigma;      // 2x2
  double **inv_xy_sigma;  // 2x2
  double **col_sigma;     // 3x3
  double **inv_col_sigma; // 3x3
} CLASS_ONE;

/*****************************************************************************/

typedef struct class_struct {
  double scale[5];
  CLASS_ONE class[MAX_CLASS];
  int n_class;
} CLASS;

/*****************************************************************************/

typedef struct ppm_data_struct {
  double **data;
  int *class;
  int n;
} PPM_DATA;

/*****************************************************************************/

extern int allocate_class_one(CLASS_ONE *c);
extern int zero_class_one(CLASS_ONE *c1);
extern int allocate_class(CLASS *c);
extern int zero_class(CLASS *c);
extern int allocate_ppm_data(PPM_DATA *m, int xdim, int ydim);
extern int convert_ppm_to_data(PPM *im, PPM_DATA *m, int allocate_flag);
extern int compute_inverse_sigmas(CLASS_ONE *c1);
extern double weighted_distance(double *v, CLASS_ONE *c1, double *scale);
extern int ppm_classify_one(double *v, CLASS *c, double *dist_out);
extern double ppm_classify_all(PPM_DATA *m, CLASS *c);
extern int ppm_update_class_one(double *v, CLASS_ONE *c1);
extern int ppm_update_class_all(PPM_DATA *m, CLASS *c);
extern int ppm_adjust_class_one(CLASS_ONE *c1);
extern int ppm_adjust_class_all(CLASS *c);
extern int draw_class_picture(RAW *im, int **colors, int n_colors, 
			      int *class, int n);
extern int draw_class_real_colors(RAW *im, CLASS *c, int *class, int n);
extern int display_class(FILE *fp, CLASS *c);
extern int threshold_color_raw(RAW *in, PPM *out, int r, int g, int b,
			       int foreground, int background);
extern int color_overlay_raw(RAW *in, PPM *t, PPM *out, int r, int g, int b);
extern int print_partition_instructions(FILE *fp);
extern int print_extraction_instructions(FILE *fp);
extern int merge_specialized(PPM *im1, PPM *im2, PPM *out);
extern int eliminate_unselected_foreground_pixels(PPM *im);
extern int display_blob2_local(FILE *fp, BLOB2 *b);
extern int extract_uniq_colors(RAW *im, int **colors);
extern int compute_class_from_blob2(CLASS *c, BLOB2 *b, int xdim, int ydim);

extern int dump_class_file(char *file, CLASS *c);
extern int read_class_file(char *file, CLASS *c, int allocate_flag);

/*****************************************************************************/

extern int collect_statistics_for_single_blob2(RAW2 *frame, PPM2 *t, 
					       int x, int y, 
					       int label, int tst,
					       BLOB2 *b);
extern int check_for_completion(PPM *t, int x, int y, int label);
